@@ -8,6 +8,7 @@ from cStringIO import StringIO
 from ConfigParser import ConfigParser
 from itertools import islice
 from shutil import rmtree
+from collections import defaultdict
 
 import tg
 from pylons import app_globals as g
@@ -46,11 +47,8 @@ class Repository(M.Repository):
             merge_request.downstream_repo_url,
         )
 
-    def count(self, branch='default'):
-        return super(Repository, self).count(branch)
-
-    def log(self, branch='default', offset=0, limit=10):
-        return super(Repository, self).log(branch, offset, limit)
+    def rev_to_commit_id(self, rev):
+        return self._impl.rev_parse(rev)
 
 class HgUI(ui.ui):
     '''Hg UI subclass that suppresses reporting of untrusted hgrc files.'''
@@ -215,23 +213,44 @@ class HgImplementation(M.RepositoryImplementation):
         doc.m.save(safe=False)
         return doc
 
-    def log(self, object_id, skip, count):
-        obj = self._hg[object_id]
-        candidates = [ obj ]
-        result = []
-        seen = set()
-        while count and candidates:
-            candidates.sort(key=lambda c:sum(c.date()))
-            obj = candidates.pop(-1)
-            if obj.hex() in seen: continue
-            seen.add(obj.hex())
-            if skip == 0:
-                result.append(obj.hex())
-                count -= 1
+    def log(self, revs=None, path=None, exclude=None, id_only=True, **kw):
+        path = path.strip('/') if path else None
+        revs = ['%s:0' % r for r in revs or []]
+        pats = ['path:%s' % path] if path else []
+        exclude = exclude or []
+        match = cmdutil.match(self._hg, pats=pats)
+        noop = lambda ctx, fns: None
+        branches = defaultdict(list)
+        for name, head in self._hg.branchtags().iteritems():
+            branches[self._hg[head].hex()].append(name)
+        change_revs = cmdutil.walkchangerevs(self._hg, match, {'rev': revs, 'prune': exclude}, noop)
+        for ctx in change_revs:
+            if id_only:
+                yield ctx.hex()
             else:
-                skip -= 1
-            candidates += self.real_parents(obj)
-        return result, [ p.hex() for p in candidates ]
+                user = ctx.user()
+                match = self.re_hg_user.match(user)
+                if match:
+                    name, email = match.groups()
+                else:
+                    name, email = (user or '--none--', '')
+                yield {
+                        'id': ctx.hex(),
+                        'message': h.really_unicode(ctx.description() or '--none--'),
+                        'authored': {
+                                'name': h.really_unicode(name),
+                                'email': h.really_unicode(email),
+                                'date': datetime.utcfromtimestamp(ctx.date()[0]),
+                            },
+                        'committed': {
+                                'name': h.really_unicode(name),
+                                'email': h.really_unicode(email),
+                                'date': datetime.utcfromtimestamp(ctx.date()[0]),
+                            },
+                        'refs': branches[ctx.hex()] + ctx.tags(),
+                        'parents': [pctx.hex() for pctx in ctx.parents() if pctx.rev() > -1],
+                        'size': int(ctx[path].size()) if path in ctx else None,
+                    }
 
     def open_blob(self, blob):
         fctx = self._hg[blob.commit._id][h.really_unicode(blob.path()).encode('utf-8')[1:]]
@@ -272,6 +291,12 @@ class HgImplementation(M.RepositoryImplementation):
             return [], []
         return [ctx.branch()], ctx.tags()
 
+    def rev_parse(self, rev):
+        try:
+            return self._hg[rev].hex()
+        except RepoLookupError as e:
+            return None
+
     def compute_tree_new(self, commit, tree_path='/'):
         ctx = self._hg[commit._id]
         fake_tree = self._tree_from_changectx(ctx)
@@ -329,6 +354,13 @@ class HgImplementation(M.RepositoryImplementation):
     def is_empty(self):
         return not self._hg or self._hg.heads() == ['\x00'*20]
 
+    def is_file(self, path, rev=None):
+        return path.strip('/') in self._hg[rev]
+
+    @LazyProperty
+    def head(self):
+        return self._hg[self._hg.tags()['tip']].hex()
+
     @LazyProperty
     def heads(self):
         return [Object(name=None, object_id=self._hg[v].hex()) for v in self._hg.heads()]
